# Copyright (C) 2016 The Qt Company Ltd.
# SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0

from dumper import Children, SubItem
from utils import TypeCode, DisplayFormat
import re

#######################################################################
#
# SSE
#
#######################################################################


def qdump____m128(d, value):
    d.putEmptyValue()
    d.putExpandable()
    if d.isExpanded():
        d.putArrayData(value.address(), 4, d.lookupType('float'))


def qdump____m256(d, value):
    d.putEmptyValue()
    d.putExpandable()
    if d.isExpanded():
        d.putArrayData(value.address(), 8, d.lookupType('float'))


def qdump____m512(d, value):
    d.putEmptyValue()
    d.putExpandable()
    if d.isExpanded():
        d.putArrayData(value.address(), 16, d.lookupType('float'))


def qdump____m128d(d, value):
    d.putEmptyValue()
    d.putExpandable()
    if d.isExpanded():
        d.putArrayData(value.address(), 2, d.lookupType('double'))


def qdump____m256d(d, value):
    d.putEmptyValue()
    d.putExpandable()
    if d.isExpanded():
        d.putArrayData(value.address(), 4, d.lookupType('double'))


def qdump____m512d(d, value):
    d.putEmptyValue()
    d.putExpandable()
    if d.isExpanded():
        d.putArrayData(value.address(), 8, d.lookupType('double'))


def qdump____m128i(d, value):
    data = d.hexencode(value.data(16))
    d.putValue(':'.join('%04x' % int(data[i:i + 4], 16) for i in range(0, 32, 4)))
    d.putExpandable()
    if d.isExpanded():
        with Children(d):
            addr = value.address()
            d.putArrayItem('uint8x16', addr, 16, 'unsigned char')
            d.putArrayItem('uint16x8', addr, 8, 'unsigned short')
            d.putArrayItem('uint32x4', addr, 4, 'unsigned int')
            d.putArrayItem('uint64x2', addr, 2, 'unsigned long long')


def qdump____m256i(d, value):
    data = d.hexencode(value.data(32))
    d.putValue(':'.join('%04x' % int(data[i:i + 4], 16) for i in range(0, 64, 4)))
    d.putExpandable()
    if d.isExpanded():
        with Children(d):
            addr = value.address()
            d.putArrayItem('uint8x32', addr, 32, 'unsigned char')
            d.putArrayItem('uint16x16', addr, 16, 'unsigned short')
            d.putArrayItem('uint32x8', addr, 8, 'unsigned int')
            d.putArrayItem('uint64x4', addr, 4, 'unsigned long long')


def qdump____m512i(d, value):
    data = d.hexencode(value.data(64))
    d.putValue(':'.join('%04x' % int(data[i:i + 4], 16) for i in range(0, 64, 4))
               + ', ' + ':'.join('%04x' % int(data[i:i + 4], 16) for i in range(64, 128, 4)))
    d.putExpandable()
    if d.isExpanded():
        with Children(d):
            d.putArrayItem('uint32x16', value.address(), 16, 'unsigned int')
            d.putArrayItem('uint64x8', value.address(), 8, 'unsigned long long')

#######################################################################
#
# GSL
#
#######################################################################


def qform__std__array():
    return [DisplayFormat.ArrayPlot]


def qdump__gsl__span(d, value):
    size, pointer = value.split('pp')
    d.putItemCount(size)
    if d.isExpanded():
        d.putPlotData(pointer, size, value.type[0])


def qdump__gsl__byte(d, value):
    d.putValue(value.integer())

#######################################################################
#
# Eigen
#
#######################################################################

#def qform__Eigen__Matrix():
#    return 'Transposed'


def qdump__Eigen__Matrix(d, value):
    innerType = value.type[0]
    argRow = value.type[1]
    argCol = value.type[2]
    options = value.type[3]
    rowMajor = (int(options) & 0x1)
    # The magic dimension value is -1 in Eigen3, but 10000 in Eigen2.
    # 10000 x 10000 matrices are rare, vectors of dim 10000 less so.
    # So 'fix' only the matrix case:
    if argCol == 10000 and argRow == 10000:
        argCol = -1
        argRow = -1
    if argCol != -1 and argRow != -1:
        nrows = argRow
        ncols = argCol
        p = value.address()
    else:
        storage = value['m_storage']
        nrows = storage['m_rows'].integer() if argRow == -1 else argRow
        ncols = storage['m_cols'].integer() if argCol == -1 else argCol
        p = storage['m_data'].pointer()
    innerSize = innerType.size()
    d.putValue('(%s x %s), %s' % (nrows, ncols, ['ColumnMajor', 'RowMajor'][rowMajor]))
    d.putField('keeporder', '1')
    d.putNumChild(nrows * ncols)

    limit = 10000
    nncols = min(ncols, limit)
    nnrows = min(nrows, limit * limit / nncols)
    if d.isExpanded():
        #format = d.currentItemFormat() # format == 1 is 'Transposed'
        with Children(d, nrows * ncols, childType=innerType):
            if ncols == 1 or nrows == 1:
                for i in range(0, min(nrows * ncols, 10000)):
                    d.putSubItem(i, d.createValue(p + i * innerSize, innerType))
            elif rowMajor == 1:
                s = 0
                for i in range(0, nnrows):
                    for j in range(0, nncols):
                        v = d.createValue(p + (i * ncols + j) * innerSize, innerType)
                        d.putNamedSubItem(s, v, '[%d,%d]' % (i, j))
                        s = s + 1
            else:
                s = 0
                for j in range(0, nncols):
                    for i in range(0, nnrows):
                        v = d.createValue(p + (i + j * nrows) * innerSize, innerType)
                        d.putNamedSubItem(s, v, '[%d,%d]' % (i, j))
                        s = s + 1


#######################################################################
#
# Nim
#
#######################################################################

def qdump__NimStringDesc(d, value):
    size, reserved = value.split('pp')
    data = value.address() + 2 * d.ptrSize()
    d.putBetterType('string')
    d.putCharArrayHelper(data, size, d.createType('char'), 'utf8')


def qdump__NimGenericSequence__(d, value, regex=r'^TY[\d]+$'):
    code = value.type.stripTypedefs().code
    if code == TypeCode.Struct:
        size, reserved = d.split('pp', value)
        data = value.address() + 2 * d.ptrSize()
        typeobj = value['data'].type.dereference()
        d.putItemCount(size)
        d.putArrayData(data, size, typeobj)
        d.putBetterType('%s (%s[%s])' % (value.type.name, typeobj.name, size))
    else:
        d.putEmptyValue()
        d.putPlainChildren(value)


def qdump__TNimNode(d, value):
    name = value['name'].pointer()
    d.putSimpleCharArray(name) if name != 0 else d.putEmptyValue()
    if d.isExpanded():
        with Children(d):
            sons = value['sons'].pointer()
            size = value['len'].integer()
            for i in range(size):
                val = d.createValue(d.extractPointer(sons + i * d.ptrSize()), value.type)
                with SubItem(d, '[%d]' % i):
                    d.putItem(val)
            with SubItem(d, '[raw]'):
                d.putPlainChildren(value)


#######################################################################
#
# D
#
#######################################################################

def cleanDType(type):
    return str(type).replace('uns long long', 'string')


def qdump_Array(d, value):
    n = value['length']
    p = value['ptr']
    t = cleanDType(value.type)[7:]
    d.putType('%s[%d]' % (t, n))
    if t == 'char':
        d.putValue(encodeCharArray(p, 100), 'local8bit')
    else:
        d.putEmptyValue()
        d.putNumChild(n)
        innerType = p.type
        if d.isExpanded():
            with Children(d, n, childType=innerType):
                for i in range(0, n):
                    d.putSubItem(i, p.dereference())
                    p = p + 1


def qdump_AArray(d, value):
    #n = value['length']
    # This ends up as _AArray_<key>_<value> with a single .ptr
    # member of type void *. Not much that can be done here.
    p = value['ptr']
    t = cleanDType(value.type)[8:]
    d.putType('%s]' % t.replace('_', '['))
    d.putEmptyValue()
    if d.isExpanded():
        with Children(d, 1):
            d.putSubItem('ptr', p)


#######################################################################
#
# MPI
#
#######################################################################

if False:

    def qdump__tree_entry(d, value):
        d.putValue('len: %s, offset: %s, type: %s' %
                   (value['blocklength'], value['offset'], value['type']))

    def qdump__tree(d, value):
        count = value['count']
        entries = value['entries']
        base = value['base'].pointer()
        d.putItemCount(count)
        if d.isExpanded():
            with Children(d):
                with SubItem(d, 'tree'):
                    d.putEmptyValue()
                    d.putNoType()
                    if d.isExpanded():
                        with Children(d):
                            for i in range(count):
                                d.putSubItem(Item(entries[i], iname))
                with SubItem(d, 'data'):
                    d.putEmptyValue()
                    d.putNoType()
                    if d.isExpanded():
                        with Children(d):
                            for i in range(count):
                                with SubItem(d, i):
                                    entry = entries[i]
                                    mpitype = str(entry['type'])
                                    d.putType(mpitype)
                                    length = int(entry['blocklength'])
                                    offset = int(entry['offset'])
                                    d.putValue('%s items at %s' % (length, offset))
                                    if mpitype == 'MPI_INT':
                                        innerType = 'int'
                                    elif mpitype == 'MPI_CHAR':
                                        innerType = 'char'
                                    elif mpitype == 'MPI_DOUBLE':
                                        innerType = 'double'
                                    else:
                                        length = 0
                                    d.putNumChild(length)
                                    if d.isExpanded():
                                        with Children(d):
                                            t = d.lookupType(innerType).pointer()
                                            p = (base + offset).cast(t)
                                            for j in range(length):
                                                d.putSubItem(j, p.dereference())


#######################################################################
#
# KDSoap
#
#######################################################################

def qdump__KDSoapValue1(d, value):
    inner = value['d']['d'].dereference()
    d.putStringValue(inner['m_name'])
    d.putPlainChildren(inner)


def qdump__KDSoapValue(d, value):
    p = (value.cast(d.lookupType('char*')) + 4).dereference().cast(d.lookupType('QString'))
    d.putStringValue(p)
    d.putPlainChildren(value['d']['d'].dereference())


#######################################################################
#
# Webkit
#
#######################################################################

def qdump__WTF__String(d, value):
    # WTF::String -> WTF::RefPtr<WTF::StringImpl> -> WTF::StringImpl*
    data = value['m_impl']['m_ptr']
    d.checkPointer(data)

    stringLength = int(data['m_length'])
    d.check(0 <= stringLength and stringLength <= 100000000)

    # WTF::StringImpl* -> WTF::StringImpl -> sizeof(WTF::StringImpl)
    offsetToData = data.type.target().size()
    bufferPtr = data.pointer() + offsetToData

    is8Bit = data['m_is8Bit']
    charSize = 1
    if not is8Bit:
        charSize = 2

    d.putCharArrayHelper(bufferPtr, stringLength, charSize)


#######################################################################
#
# Python
#
#######################################################################

def get_python_interpreter_major_version(d):
    key = 'python_interpreter_major_version'
    if key in d.generalCache:
        return d.generalCache[key]

    e = "(char*)Py_GetVersion()"
    result = d.nativeParseAndEvaluate(e)
    result_str = str(result)
    matches = re.search(r'(\d+?)\.(\d+?)\.(\d+?)', result_str)
    if matches:
        result_str = matches.group(1)
    d.generalCache[key] = result_str
    return result_str


def is_python_3(d):
    return get_python_interpreter_major_version(d) == '3'


def repr_cache_decorator(namespace):
    def real_decorator(func_to_decorate):
        def wrapper(d, address):
            if namespace in d.perStepCache and address in d.perStepCache[namespace]:
                return d.perStepCache[namespace][address]

            if namespace not in d.perStepCache:
                d.perStepCache[namespace] = {}

            if address == 0:
                result_str = d.perStepCache[namespace][address] = "<nullptr>"
                return result_str

            result = func_to_decorate(d, address)
            d.perStepCache[namespace][address] = result
            return result
        return wrapper
    return real_decorator


@repr_cache_decorator('py_object')
def get_py_object_repr_helper(d, address):
    # The code below is a long way to evaluate:
    # ((PyBytesObject *)PyUnicode_AsEncodedString(PyObject_Repr(
    #        (PyObject*){}), \"utf-8\", \"backslashreplace\"))->ob_sval"
    # But with proper object cleanup.

    e_decref = "Py_DecRef((PyObject *){})"

    e = "PyObject_Repr((PyObject*){})"
    repr_object_value = d.parseAndEvaluate(e.format(address))
    repr_object_address = d.fromPointerData(repr_object_value.ldata)[0]

    if is_python_3(d):
        # Try to get a UTF-8 encoded string from the repr object.
        e = "PyUnicode_AsEncodedString((PyObject*){}, \"utf-8\", \"backslashreplace\")"
        string_object_value = d.parseAndEvaluate(e.format(repr_object_address))
        string_object_address = d.fromPointerData(string_object_value.ldata)[0]

        e = "(char*)(((PyBytesObject *){})->ob_sval)"
        result = d.nativeParseAndEvaluate(e.format(string_object_address))

        # It's important to stringify the result before any other evaluations happen.
        result_str = str(result)

        # Clean up.
        d.nativeParseAndEvaluate(e_decref.format(string_object_address))

    else:
        # Retrieve non-unicode string.
        e = "(char*)(PyString_AsString((PyObject*){}))"
        result = d.nativeParseAndEvaluate(e.format(repr_object_address))

        # It's important to stringify the result before any other evaluations happen.
        result_str = str(result)

    # Do some string stripping.
    # FIXME when using cdb engine.
    matches = re.search(r'.+?"(.+)"$', result_str)
    if matches:
        result_str = matches.group(1)

    # Clean up.
    d.nativeParseAndEvaluate(e_decref.format(repr_object_address))

    return result_str


@repr_cache_decorator('py_object_type')
def get_py_object_type(d, object_address):
    e = "((PyObject *){})->ob_type"
    type_value = d.parseAndEvaluate(e.format(object_address))
    type_address = d.fromPointerData(type_value.ldata)[0]
    type_repr = get_py_object_repr_helper(d, type_address)
    return type_repr


@repr_cache_decorator('py_object_meta_type')
def get_py_object_meta_type(d, object_address):
    # The python3 object layout has a few more indirections.
    if is_python_3(d):
        e = "((PyObject *){})->ob_type->ob_base->ob_base->ob_type"
    else:
        e = "((PyObject *){})->ob_type->ob_type"
    type_value = d.parseAndEvaluate(e.format(object_address))
    type_address = d.fromPointerData(type_value.ldata)[0]
    type_repr = get_py_object_repr_helper(d, type_address)
    return type_repr


@repr_cache_decorator('py_object_base_class')
def get_py_object_base_class(d, object_address):
    e = "((PyObject *){})->ob_type->tp_base"
    base_value = d.parseAndEvaluate(e.format(object_address))
    base_address = d.fromPointerData(base_value.ldata)[0]
    base_repr = get_py_object_repr_helper(d, base_address)
    return base_repr


def get_py_object_repr(d, value):
    address = value.address()
    repr_available = False
    try:
        result = get_py_object_repr_helper(d, address)
        d.putValue(d.hexencode(result), encoding='utf8')
        repr_available = True
    except:
        d.putEmptyValue()

    def sub_item(name, functor, address):
        with SubItem(d, '[{}]'.format(name)):
            sub_value = functor(d, address)
            d.putValue(d.hexencode(sub_value), encoding='utf8')

    d.putExpandable()
    if d.isExpanded():
        with Children(d):
            if repr_available:
                sub_item('class', get_py_object_type, address)
                sub_item('super class', get_py_object_base_class, address)
                sub_item('meta type', get_py_object_meta_type, address)
            d.putFields(value)


def qdump__PyTypeObject(d, value):
    get_py_object_repr(d, value)


def qdump___typeobject(d, value):
    get_py_object_repr(d, value)


def qdump__PyObject(d, value):
    get_py_object_repr(d, value)


def qdump__PyVarObject(d, value):
    get_py_object_repr(d, value)


#######################################################################
#
# Internal test
#
#######################################################################

def qdump__QtcDumperTest_FieldAccessByIndex(d, value):
    d.putValue(value["d"][2].integer())


def qdump__QtcDumperTest_PointerArray(d, value):
    foos = value["foos"]
    d.putItemCount(10)
    if d.isExpanded():
        with Children(d, 10):
            for i in d.childRange():
                d.putSubItem(i, foos[i])


def qdump__QtcDumperTest_BufArray(d, value):
    maxItems = 1000
    buffer = value['buffer']
    count = int(value['count'])
    objsize = int(value['objSize'])
    valueType = value.type.templateArgument(0)
    d.putItemCount(count, maxItems)
    d.putNumChild(count)
    if d.isExpanded():
        with Children(d, count, maxNumChild=maxItems, childType=valueType):
            for i in d.childRange():
                d.putSubItem(i, (buffer + (i * objsize)).dereference().cast(valueType))


def qdump__QtcDumperTest_List__NodeX(d, value):
    typename = value.type.unqualified().name
    pos0 = typename.find('<')
    pos1 = typename.find('>')
    tName = typename[pos0 + 1:pos1]
    d.putBetterType('QtcDumperTest_List<' + tName + '>::Node')
    d.putExpandable()
    if d.isExpanded():
        obj_type = d.lookupType(tName)
        with Children(d):
            d.putSubItem("this", value.cast(obj_type))
            d.putFields(value)
    #d.putPlainChildren(value)


def qdump__QtcDumperTest_List(d, value):
    innerType = value.type[0]
    d.putExpandable()
    p = value['root']
    if d.isExpanded():
        with Children(d):
            d.putSubItem("[p]", p)
            d.putSubItem("[root]", value["root"].cast(innerType))
            d.putFields(value)
    #d.putPlainChildren(value)


def qdump__QtcDumperTest_String(d, value):
    with Children(d):
        first = d.hexdecode(d.putSubItem('first', value['first']).value)
        second = d.hexdecode(d.putSubItem('second', value['second']).value)
        third = d.hexdecode(d.putSubItem('third', value['third']).value)[:-1]
    d.putValue(first + ', ' + second + ', ' + third)



# Copyright (C) 2022 The Qt Company Ltd.
# SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0

from stdtypes import qdump__std__array, qdump__std__complex, qdump__std__once_flag, qdump__std__unique_ptr, qdumpHelper__std__deque__libcxx, qdumpHelper__std__vector__libcxx
from utils import DisplayFormat
from dumper import Children, DumperBase


def qform__std____1__array():
    return [DisplayFormat.ArrayPlot]


def qdump__std____1__array(d, value):
    qdump__std__array(d, value)


def qdump__std____1__complex(d, value):
    qdump__std__complex(d, value)


def qdump__std____1__deque(d, value):
    qdumpHelper__std__deque__libcxx(d, value)


def qdump__std____1__list(d, value):
    if value.type.size() == 3 * d.ptrSize():
        # C++11 only.
        (dummy1, dummy2, size) = value.split("ppp")
        d.putItemCount(size)
    else:
        # Need to count manually.
        p = d.extractPointer(value)
        head = value.address()
        size = 0
        while head != p and size < 1001:
            size += 1
            p = d.extractPointer(p)
        d.putItemCount(size, 1000)

    if d.isExpanded():
        (prev, p) = value.split("pp")
        innerType = value.type[0]
        typeCode = "pp{%s}" % innerType.name
        with Children(d, size, maxNumChild=1000, childType=innerType):
            for i in d.childRange():
                (prev, p, val) = d.split(typeCode, p)
                d.putSubItem(i, val)


def qdump__std____1__set(d, value):
    (proxy, head, size) = value.split("ppp")

    d.check(0 <= size and size <= 100 * 1000 * 1000)
    d.putItemCount(size)

    if d.isExpanded():
        valueType = value.type[0]

        def in_order_traversal(node):
            (left, right, parent, color, pad, data) = d.split("pppB@{%s}" % (valueType.name), node)

            if left:
                for res in in_order_traversal(left):
                    yield res

            yield data

            if right:
                for res in in_order_traversal(right):
                    yield res

        with Children(d, size, maxNumChild=1000):
            for (i, data) in zip(d.childRange(), in_order_traversal(head)):
                d.putSubItem(i, data)


def qdump__std____1__multiset(d, value):
    qdump__std____1__set(d, value)


def qform__std____1__map():
    return [DisplayFormat.CompactMap]


def qdump__std____1__map(d, value):
    try:
        (proxy, head, size) = value.split("ppp")
        d.check(0 <= size and size <= 100 * 1000 * 1000)

    # Sometimes there is extra data at the front. Don't know why at the moment.
    except RuntimeError:
        (junk, proxy, head, size) = value.split("pppp")
        d.check(0 <= size and size <= 100 * 1000 * 1000)

    d.putItemCount(size)

    if d.isExpanded():
        keyType = value.type[0]
        valueType = value.type[1]
        pairType = value.type[3][0]

        def in_order_traversal(node):
            (left, right, parent, color, pad, pair) = d.split("pppB@{%s}" % (pairType.name), node)

            if left:
                for res in in_order_traversal(left):
                    yield res

            yield pair.split("{%s}@{%s}" % (keyType.name, valueType.name))[::2]

            if right:
                for res in in_order_traversal(right):
                    yield res

        with Children(d, size, maxNumChild=1000):
            for (i, pair) in zip(d.childRange(), in_order_traversal(head)):
                d.putPairItem(i, pair)


def qform__std____1__multimap():
    return [DisplayFormat.CompactMap]


def qdump__std____1__multimap(d, value):
    qdump__std____1__map(d, value)


def qdump__std____1__map__iterator(d, value):
    d.putEmptyValue()
    if d.isExpanded():
        with Children(d):
            node = value['__i_']['__ptr_'].dereference()['__value_']['__cc']
            d.putSubItem('first', node['first'])
            d.putSubItem('second', node['second'])


def qdump__std____1__map__const_iterator(d, value):
    qdump__std____1__map__iterator(d, value)


def qdump__std____1__set__iterator(d, value):
    d.putEmptyValue()
    d.putExpandable()
    if value.type.name.endswith("::iterator"):
        treeTypeName = value.type.name[:-len("::iterator")]
    elif value.type.name.endswith("::const_iterator"):
        treeTypeName = value.type.name[:-len("::const_iterator")]
    treeType = d.lookupType(treeTypeName)
    keyType = treeType[0]
    if d.isExpanded():
        with Children(d):
            node = value['__ptr_'].dereference()['__value_']
            node = node.cast(keyType)
            d.putSubItem('value', node)


def qdump__std____1__set_const_iterator(d, value):
    qdump__std____1__set__iterator(d, value)


def qdump__std____1__stack(d, value):
    d.putItem(value["c"])
    d.putBetterType(value.type)


def GetChildMemberWithName(value, name):
    members = value.members(True)

    for member in members:
        if member.name == name:
            return member
    return None


def GetIndexOfChildWithName(value, name):
    members = value.members(True)

    for i, member in enumerate(members):
        if member.name == name:
            return i
    return None


class StringLayout:
    CSD = 0
    DSC = 1


def std_1_string_dumper_v2(d, value):
    charType = value['__l']['__data_'].dereference().type

    R = GetChildMemberWithName(value, "__r_")
    if not R:
        raise Exception("Could not find __r_")

    # __r_ is a compressed_pair of the actual data and the allocator. The data we
    # want is in the first base class.
    R_Base_SP = R[0]

    if not R_Base_SP:
        raise Exception("Could not find R_Base_SP")

    Rep_Sp = GetChildMemberWithName(R_Base_SP, "__value_")

    if not Rep_Sp:
        raise Exception("Could not find __value_")

    # Our layout seems a little different
    Rep_Sp = Rep_Sp[0]

    if not Rep_Sp:
        raise Exception("Could not find Rep_Sp")

    L = GetChildMemberWithName(Rep_Sp, "__l")

    if not L:
        raise Exception("Could not find __l")

    layout = StringLayout.CSD
    if GetIndexOfChildWithName(L, "__data_") == 0:
        layout = StringLayout.DSC

    short_mode = False
    using_bitmasks = True
    size = 0
    size_mode_value = 0

    Short_Sp = GetChildMemberWithName(Rep_Sp, "__s")
    if not Short_Sp:
        raise Exception("Could not find __s")

    Is_Long = GetChildMemberWithName(Short_Sp, "__is_long_")
    Size_Sp = GetChildMemberWithName(Short_Sp, "__size_")
    if not Size_Sp:
        raise Exception("Could not find __size_")

    if Is_Long:
        using_bitmasks = False
        short_mode = Is_Long.integer() == 0
        size = Size_Sp.integer()
    else:
        size_mode_value = Size_Sp.integer()
        mode_mask = 1
        if layout == StringLayout.DSC:
            mode_mask = 0x80
        short_mode = (size_mode_value & mode_mask) == 0

    if short_mode:
        Location_Sp = GetChildMemberWithName(Short_Sp, "__data_")

        if using_bitmasks:
            size = ((size_mode_value >> 1) % 256)
            if layout == StringLayout.DSC:
                size = size_mode_value

        # The string is most likely not initialized yet
        if size > 100 or not Location_Sp:
            raise Exception("Probably not initialized yet")

        d.putCharArrayHelper(d.extractPointer(Location_Sp), size,
                             charType, d.currentItemFormat())
        return

    Location_Sp = GetChildMemberWithName(L, "__data_")
    Size_Vo = GetChildMemberWithName(L, "__size_")
    Capacity_Vo = GetChildMemberWithName(L, "__cap_")

    if not Location_Sp or not Size_Vo or not Capacity_Vo:
        raise Exception("Could not find Location_Sp, Size_Vo or Capacity_Vo")

    size = Size_Vo.integer()
    capacity = Capacity_Vo.integer()
    if not using_bitmasks and layout == StringLayout.CSD:
        capacity *= 2
    if capacity < size:
        raise Exception("Capacity is less than size")

    d.putCharArrayHelper(d.extractPointer(Location_Sp), size,
                         charType, d.currentItemFormat())


def std_1_string_dumper_v1(d, value):
    charType = value['__l']['__data_'].dereference().type
    D = None

    if d.isLldb:
        D = value[0][0][0][0]
    elif d.isGdb:
        D = value["__r_"].members(True)[0][0][0]
    else:
        raise Exception("Unknown debugger (neither gdb nor lldb)")

    layoutDecider = D[0][0]
    if not layoutDecider:
        raise Exception("Could not find layoutDecider")

    size = 0
    size_mode_value = 0
    short_mode = False
    libcxx_version = 14

    layoutModeIsDSC = layoutDecider.name == '__data_'
    if (layoutModeIsDSC):
        size_mode = D[1][1][0]
        if not size_mode:
            raise Exception("Could not find size_mode")
        if not size_mode.name == '__size_':
            size_mode = D[1][1][1]
            if not size_mode:
                raise Exception("Could not find size_mode")

        size_mode_value = size_mode.integer()
        short_mode = ((size_mode_value & 0x80) == 0)
    else:
        size_mode = D[1][0][0]
        if not size_mode:
            raise Exception("Could not find size_mode")

        if size_mode.name == '__is_long_':
            libcxx_version = 15
            short_mode = (size_mode.integer() == 0)

            size_mode = D[1][0][1]
            size_mode_value = size_mode.integer()
        else:
            size_mode_value = size_mode.integer()
            short_mode = ((size_mode_value & 1) == 0)

    if short_mode:
        s = D[1]

        if not s:
            raise Exception("Could not find s")

        if libcxx_version == 14:
            location_sp = s[0] if layoutModeIsDSC else s[1]
            size = size_mode_value if layoutModeIsDSC else ((size_mode_value >> 1) % 256)
        elif libcxx_version == 15:
            location_sp = s[0] if layoutModeIsDSC else s[2]
            size = size_mode_value

    else:
        l = D[0]
        if not l:
            raise Exception("Could not find l")

        # we can use the layout_decider object as the data pointer
        location_sp = layoutDecider if layoutModeIsDSC else l[2]
        size_vo = l[1]
        if not size_vo or not location_sp:
            raise Exception("Could not find size_vo or location_sp")
        size = size_vo.integer()

    if short_mode and location_sp:
        d.putCharArrayHelper(d.extractPointer(location_sp), size,
                             charType, d.currentItemFormat())
    else:
        d.putCharArrayHelper(location_sp.integer(),
                             size, charType, d.currentItemFormat())

    return

def qdump__std____1__string(d, value):
    try:
        std_1_string_dumper_v2(d, value)
    except Exception as eV2:
        try:
            std_1_string_dumper_v1(d, value)
        except Exception as eV1:
            d.putValue("Could not parse: %s, %s" % (eV1, eV2))


def qdump__std____1__wstring(d, value):
    try:
        std_1_string_dumper_v2(d, value)
    except Exception as eV2:
        try:
            std_1_string_dumper_v1(d, value)
        except Exception as eV1:
            d.putValue("Could not parse: %s, %s" % (eV1, eV2))


def qdump__std____1__basic_string(d, value):
    innerType = value.type[0].name
    if innerType in ("char", "char8_t", "char16_t"):
        qdump__std____1__string(d, value)
    elif innerType in ("wchar_t", "char32_t"):
        qdump__std____1__wstring(d, value)
    else:
        d.warn("UNKNOWN INNER TYPE %s" % innerType)


def qdump__std____1__shared_ptr(d, value):
    i = value["__ptr_"]
    if i.pointer() == 0:
        d.putValue("(null)")
    else:
        d.putItem(i.dereference())
        d.putBetterType(value.type)


def qdump__std____1__weak_ptr(d, value):
    return qdump__std____1__shared_ptr(d, value)


def qdump__std____1__unique_ptr(d, value):
    if value.type.size() == d.ptrSize():
        p = d.extractPointer(value)
    else:
        _, p = value.split("pp"); # For custom deleters.
    if p == 0:
        d.putValue("(null)")
    else:
        try:
            d.putItem(value["__value_"])
            d.putValue(d.currentValue.value, d.currentValue.encoding)
        except:
            d.putItem(d.createValue(p, value.type[0]))
    d.putBetterType(value.type)


def qform__std____1__unordered_map():
    return [DisplayFormat.CompactMap]


def qdump__std____1__unordered_map(d, value):
    (size, _) = value["__table_"]["__p2_"].split("pp")
    d.putItemCount(size)

    keyType = value.type[0]
    valueType = value.type[1]
    pairType = value.type[4][0]

    if d.isExpanded():
        curr = value["__table_"]["__p1_"].split("pp")[0]

        def traverse_list(node):
            while node:
                (next_, _, pad, pair) = d.split("pp@{%s}" % (pairType.name), node)
                yield pair.split("{%s}@{%s}" % (keyType.name, valueType.name))[::2]
                node = next_

        with Children(d, size, childType=value.type[0], maxNumChild=1000):
            for (i, value) in zip(d.childRange(), traverse_list(curr)):
                d.putPairItem(i, value, 'key', 'value')


def qdump__std____1__unordered_set(d, value):
    (size, _) = value["__table_"]["__p2_"].split("pp")
    d.putItemCount(size)

    valueType = value.type[0]

    if d.isExpanded():
        curr = value["__table_"]["__p1_"].split("pp")[0]

        def traverse_list(node):
            while node:
                (next_, _, pad, val) = d.split("pp@{%s}" % (valueType.name), node)
                yield val
                node = next_

        with Children(d, size, childType=value.type[0], maxNumChild=1000):
            for (i, value) in zip(d.childRange(), traverse_list(curr)):
                d.putSubItem(i, value)


def qdump__std____1__unordered_multiset(d, value):
    qdump__std____1__unordered_set(d, value)


def qform__std____1__valarray():
    return [DisplayFormat.ArrayPlot]


def qdump__std____1__valarray(d, value):
    innerType = value.type[0]
    (begin, end) = value.split('pp')
    size = int((end - begin) / innerType.size())
    d.putItemCount(size)
    d.putPlotData(begin, size, innerType)


def qform__std____1__vector():
    return [DisplayFormat.ArrayPlot]


def qdump__std____1__vector(d, value):
    qdumpHelper__std__vector__libcxx(d, value)


def qdump__std____1__once_flag(d, value):
    qdump__std__once_flag(d, value)


def qdump__std____1__variant(d, value):
    index = value['__impl']['__index']
    index_num = int(index)
    value_type = d.templateArgument(value.type, index_num)
    d.putValue("<%s:%s>" % (index_num, value_type.name))

    d.putNumChild(2)
    if d.isExpanded():
        with Children(d):
            d.putSubItem("index", index)
            d.putSubItem("value", value.cast(value_type))


def qdump__std____1__optional(d, value):
    if value['__engaged_'].integer() == 0:
        d.putSpecialValue("empty")
    else:
        d.putItem(value['#1']['__val_'])


def qdump__std____1__tuple(d, value):
    values = []
    for member in value['__base_'].members(False):
        values.append(member['__value_'])
    d.putItemCount(len(values))
    d.putNumChild(len(values))
    if d.isExpanded():
        with Children(d):
            count = 0
            for internal_value in values:
                d.putSubItem("[%i]" % count, internal_value)
                count += 1
